/* linux/include/asm/arch-s3c2410/regs-clock.h
 *
 * Copyright (c) 2003,2004 Simtec Electronics <linux@simtec.co.uk>
 *		      http://www.simtec.co.uk/products/SWLINUX/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * S3C2410 clock register definitions
 *
 *  Changelog:
 *    18-Aug-2004 Ben Dooks         Added 2440 definitions
 *    08-Aug-2004 Herbert Ptzl     Added CLKCON definitions
 *    19-06-2003  Ben Dooks         Created file
 *    12-03-2004  Ben Dooks         Updated include protection
 *    29-Sep-2004 Ben Dooks	    Fixed usage for assembly inclusion
 *    10-Feb-2005 Ben Dooks	    Fixed CAMDIVN address (Guillaume Gourat)
 *    10-Mar-2005 Lucas Villa Real  Changed S3C2410_VA to S3C24XX_VA
 *    28-Feb-2006 Koen Martens      Added s3c2412_get_pll
 */

#ifndef __ASM_ARM_REGS_CLOCK
#define __ASM_ARM_REGS_CLOCK "$Id: clock.h,v 1.4 2003/04/30 14:50:51 ben Exp $"

#define S3C2410_CLKREG(x) ((x) + S3C24XX_VA_CLKPWR)

#define S3C2410_PLLVAL(_m,_p,_s) ((_m) << 12 | ((_p) << 4) | ((_s)))

#define S3C2410_LOCKTIME    S3C2410_CLKREG(0x00)
#define S3C2410_MPLLCON	    S3C2410_CLKREG(0x04)
#define S3C2410_UPLLCON	    S3C2410_CLKREG(0x08)
#define S3C2410_CLKCON	    S3C2410_CLKREG(0x0C)
#define S3C2410_CLKSLOW	    S3C2410_CLKREG(0x10)
#define S3C2410_CLKDIVN	    S3C2410_CLKREG(0x14)

#define S3C2410_CLKCON_IDLE	     (1<<2)
#define S3C2410_CLKCON_POWER	     (1<<3)
#define S3C2410_CLKCON_NAND	     (1<<4)
#define S3C2410_CLKCON_LCDC	     (1<<5)
#define S3C2410_CLKCON_USBH	     (1<<6)
#define S3C2410_CLKCON_USBD	     (1<<7)
#define S3C2412_CLKCON_USBD	     (1<<16)
#define S3C2410_CLKCON_PWMT	     (1<<8)
#define S3C2410_CLKCON_SDI	     (1<<9)
#define S3C2410_CLKCON_UART0	     (1<<10)
#define S3C2410_CLKCON_UART1	     (1<<11)
#define S3C2410_CLKCON_UART2	     (1<<12)
#define S3C2410_CLKCON_GPIO	     (1<<13)
#define S3C2410_CLKCON_RTC	     (1<<14)
#define S3C2410_CLKCON_ADC	     (1<<15)
#define S3C2410_CLKCON_IIC	     (1<<16)
#define S3C2410_CLKCON_IIS	     (1<<17)
#define S3C2410_CLKCON_SPI	     (1<<18)

#define S3C2410_PLLCON_MDIVSHIFT     12
#define S3C2410_PLLCON_PDIVSHIFT     4
#define S3C2410_PLLCON_SDIVSHIFT     0
#define S3C2410_PLLCON_MDIVMASK	     ((1<<(1+(19-12)))-1)
#define S3C2410_PLLCON_PDIVMASK	     ((1<<5)-1)
#define S3C2410_PLLCON_SDIVMASK	     3

/* DCLKCON register addresses in gpio.h */

#define S3C2410_DCLKCON_DCLK0EN	     (1<<0)
#define S3C2410_DCLKCON_DCLK0_PCLK   (0<<1)
#define S3C2410_DCLKCON_DCLK0_UCLK   (1<<1)
#define S3C2410_DCLKCON_DCLK0_DIV(x) (((x) - 1 )<<4)
#define S3C2410_DCLKCON_DCLK0_CMP(x) (((x) - 1 )<<8)

#define S3C2410_DCLKCON_DCLK1EN	     (1<<16)
#define S3C2410_DCLKCON_DCLK1_PCLK   (0<<17)
#define S3C2410_DCLKCON_DCLK1_UCLK   (1<<17)
#define S3C2410_DCLKCON_DCLK1_DIV(x) (((x) - 1) <<20)

#define S3C2410_CLKDIVN_PDIVN	     (1<<0)
#define S3C2410_CLKDIVN_HDIVN	     (1<<1)

#ifndef __ASSEMBLY__

static inline unsigned int
s3c2410_get_pll(int pllval, int baseclk)
{
	int mdiv, pdiv, sdiv;

	mdiv = pllval >> S3C2410_PLLCON_MDIVSHIFT;
	pdiv = pllval >> S3C2410_PLLCON_PDIVSHIFT;
	sdiv = pllval >> S3C2410_PLLCON_SDIVSHIFT;

	mdiv &= S3C2410_PLLCON_MDIVMASK;
	pdiv &= S3C2410_PLLCON_PDIVMASK;
	sdiv &= S3C2410_PLLCON_SDIVMASK;

	return (baseclk * (mdiv + 8)) / ((pdiv + 2) << sdiv);
}

#ifdef CONFIG_CPU_S3C2412
static inline unsigned int
s3c2412_get_pll(int pllval, int baseclk)
{
	int mdiv, pdiv, sdiv;

	mdiv = pllval >> S3C2410_PLLCON_MDIVSHIFT;
	pdiv = pllval >> S3C2410_PLLCON_PDIVSHIFT;
	sdiv = pllval >> S3C2410_PLLCON_SDIVSHIFT;

	mdiv &= S3C2410_PLLCON_MDIVMASK;
	pdiv &= S3C2410_PLLCON_PDIVMASK;
	sdiv &= S3C2410_PLLCON_SDIVMASK;

	return (baseclk * ((mdiv + 8)<<1) ) / ((pdiv + 2) << sdiv);
}
#endif /* CONFIG_CPU_S3C2412 */

#endif /* __ASSEMBLY__ */

#ifdef CONFIG_CPU_S3C2440

/* extra registers */
#define S3C2440_CAMDIVN	    S3C2410_CLKREG(0x18)

#define S3C2440_CLKCON_CAMERA        (1<<19)
#define S3C2440_CLKCON_AC97          (1<<20)

#define S3C2440_CLKDIVN_PDIVN	     (1<<0)
#define S3C2440_CLKDIVN_HDIVN_MASK   (3<<1)
#define S3C2440_CLKDIVN_HDIVN_1      (0<<1)
#define S3C2440_CLKDIVN_HDIVN_2      (1<<1)
#define S3C2440_CLKDIVN_HDIVN_4_8    (2<<1)
#define S3C2440_CLKDIVN_HDIVN_3_6    (3<<1)
#define S3C2440_CLKDIVN_UCLK         (1<<3)

#define S3C2440_CAMDIVN_CAMCLK_MASK  (0xf<<0)
#define S3C2440_CAMDIVN_CAMCLK_SEL   (1<<4)
#define S3C2440_CAMDIVN_HCLK3_HALF   (1<<8)
#define S3C2440_CAMDIVN_HCLK4_HALF   (1<<9)
#define S3C2440_CAMDIVN_DVSEN        (1<<12)

#endif /* CONFIG_CPU_S3C2440 */

#ifdef CONFIG_CPU_S3C2412

#define S3C2412_CLKCON_DMA0          (1<<0)
#define S3C2412_CLKCON_DMA1          (1<<1)
#define S3C2412_CLKCON_DMA2          (1<<2)
#define S3C2412_CLKCON_DMA3          (1<<3)
#define S3C2412_CLKCON_NAND          (1<<4)
#define S3C2412_CLKCON_LCDC          (1<<5)
#define S3C2412_CLKCON_USBH          (1<<6)
#define S3C2412_CLKCON_SDRAM         (1<<8)
#define S3C2412_CLKCON_HCLKx1_2      (1<<10)
#define S3C2412_CLKCON_USBD48M       (1<<11)
#define S3C2412_CLKCON_USBH48M       (1<<12)
#define S3C2412_CLKCON_IISCLK        (1<<13)
#define S3C2412_CLKCON_UARTCLK       (1<<14)
#define S3C2412_CLKCON_USBD          (1<<16)
#define S3C2412_CLKCON_PWMT          (1<<17)
#define S3C2412_CLKCON_SDI           (1<<18)
#define S3C2412_CLKCON_UART0         (1<<19)
#define S3C2412_CLKCON_UART1         (1<<20)
#define S3C2412_CLKCON_UART2         (1<<21)
#define S3C2412_CLKCON_GPIO          (1<<22)
#define S3C2412_CLKCON_RTC           (1<<23)
#define S3C2412_CLKCON_ADC           (1<<24)
#define S3C2412_CLKCON_IIC           (1<<25)
#define S3C2412_CLKCON_IIS           (1<<26)
#define S3C2412_CLKCON_SPI           (1<<27)
#define S3C2412_CLKCON_WDT           (1<<28)

#define S3C2412_CLKDIVN_HCLKDIV_MASK     (3<<0)
#define S3C2412_CLKDIVN_HCLKDIV_1_2      (0<<0)
#define S3C2412_CLKDIVN_HCLKDIV_2_4      (1<<0)
#define S3C2412_CLKDIVN_HCLKDIV_3_6      (2<<0)
#define S3C2412_CLKDIVN_HCLKDIV_4_8      (3<<0)
#define S3C2412_CLKDIVN_PCLKDIV          (1<<2)
#define S3C2412_CLKDIVN_ARMDIV           (1<<3)
#define S3C2412_CLKDIVN_DVSEN            (1<<4)
#define S3C2412_CLKDIVN_HALFHCLK         (1<<5)
#define S3C2412_CLKDIVN_USB48DIV         (1<<6)
#define S3C2412_CLKDIVN_UARTCLKDIV_MASK  (0xf<<8)
#define S3C2412_CLKDIVN_IISCLKDIV_MASK   (0xf<<12)

#define S3C2412_OSCSET                   S3C2410_CLKREG(0x18)
#define S3C2412_CLKSRC                   S3C2410_CLKREG(0x1c)
#define S3C2412_PMCON                    S3C2410_CLKREG(0x20)
#define S3C2412_PWRCFG                   S3C2410_CLKREG(0x24)
#define S3C2412_ENDIAN                   S3C2410_CLKREG(0x2c)
#define S3C2412_SWRSTCON                 S3C2410_CLKREG(0x30)
#define S3C2412_RSTCON                   S3C2410_CLKREG(0x34)
#define S3C2412_INFORM0                  S3C2410_CLKREG(0x70)
#define S3C2412_INFORM1                  S3C2410_CLKREG(0x74)
#define S3C2412_INFORM2                  S3C2410_CLKREG(0x78)
#define S3C2412_INFORM3                  S3C2410_CLKREG(0x7c)

#define S3C2412_OSCSET_XTALWAIT_MASK     (0xffff<<0)

#define S3C2412_CLKSRC_EXTCLKDIV_MASK    (7<<0)
#define S3C2412_CLKSRC_SELEXTDIV         (1<<3)
#define S3C2412_CLKSRC_SELMPLL           (1<<4)
#define S3C2412_CLKSRC_SELUPLL           (1<<5)
#define S3C2412_CLKSRC_SELUART           (1<<8)
#define S3C2412_CLKSRC_SELIIS            (1<<9)
#define S3C2412_CLKSRC_SELUSB            (1<<10)
#define S3C2412_CLKSRC_SELUREF_MASK      (3<<12)
#define S3C2412_CLKSRC_SELUREF_OM4       (0<<12)
#define S3C2412_CLKSRC_SELUREF_EXTOSC    (2<<12)
#define S3C2412_CLKSRC_SELUREF_EXTCLK    (3<<12)
#define S3C2412_CLKSRC_SELEREF_MASK      (3<<14)
#define S3C2412_CLKSRC_SELEREF_OM4       (0<<12)
#define S3C2412_CLKSRC_SELEREF_EXTOSC    (2<<12)
#define S3C2412_CLKSRC_SELEREF_EXTCLK    (3<<12)

#define S3C2412_PMCON_MODESTOP           (1<<16)
#define S3C2412_PMCON_MODESLEEP_MASK     (0xffff<<0)
#define S3C2412_PMCON_GOTOBED            (0x2bed)

#define S3C2412_PWRCFG_BATF_CFG_MASK     (3<<0)
#define S3C2412_PWRCFG_BATF_CFG_IGN      (0<<0)
#define S3C2412_PWRCFG_BATF_CFG_SLEEP    (2<<0)
#define S3C2412_PWRCFG_STBYWFI_MASK      (3<<6)
#define S3C2412_PWRCFG_STBYWFI_IGN       (0<<6)
#define S3C2412_PWRCFG_STBYWFI_IDLE      (1<<6)
#define S3C2412_PWRCFG_STBYWFI_STOP      (2<<6)
#define S3C2412_PWRCFG_STBYWFI_SLEEP     (3<<6)
#define S3C2412_PWRCFG_RTC_CFG           (1<<8)
#define S3C2412_PWRCFG_NFRESET_CFG       (1<<9)
#define S3C2412_PWRCFG_ESLEEP_CFG        (1<<15)
#define S3C2412_PWRCFG_EINT_WAKE_MASK    (0xffff<<16)

#ifndef __ASSEMBLY__

static inline void s3c2412_gen_stbywfi(unsigned cfg)
{
	unsigned reg;
	const int zero = 0;

	reg = __raw_readl(S3C2412_PWRCFG);
	reg &= ~S3C2412_PWRCFG_STBYWFI_MASK;
	reg |= cfg;
	__raw_writel(reg, S3C2412_PWRCFG);

	asm("mcr%? p15, 0, %0, c7, c0, 4" : : "r" (zero));
}

#endif /* __ASSEMBLY__ */

#define S3C2412_SWRSTCON_SWRST           (0x533c2412)

#define S3C2412_RSTCON_PWRSETCNT_MASK    (0xff<<0)
#define S3C2412_RSTCON_RSTCNT_MASK       (0xff<<8)

#define S3C2412_ENDIAN_BIG               (1<<0)

#endif /* CONFIG_CPU_S3C2412 */

#endif /* __ASM_ARM_REGS_CLOCK */
